{
   MegaBrutal's SMTP Server (MgSMTP)
   Copyright (C) 2010-2018 MegaBrutal

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

{
   Unit: Log
   Supports basic logging features.
}


{$MODE DELPHI}
unit Log;

interface
uses SysUtils, Classes, INIFiles, Common;

type

   TStreamLogger = class
      constructor Create(Stream: TStream);
      destructor Destroy; override;
   protected
      LogFile: TStream;
      function AddPureLine(Line: string): boolean; virtual;
   public
      function AddLine(Line: string): boolean; overload;
      function AddLine(Agent, Line: string): boolean; overload;
      function WriteLn(Line: string): boolean; overload;
      function WriteLn: boolean; overload;
   end;


   TLogger = class(TStreamLogger)
      constructor Create(Config: TINIFile);
      destructor Destroy; override;
   protected
      FLogOn: boolean;
      FLogFileName: string;
      {FLogLevel: integer;}
      function AddPureLine(Line: string): boolean; override;
   public
      function AddStdLine(Line: string): boolean; overload;
      function AddStdLine(Agent, Line: string): boolean; overload;
   end;


var

   Out: TStreamLogger;
   Logger: TLogger;



implementation

const

   LOG_FN_STDOUT        =  'mgsmtp_stdout.log';
   LOG_FN_STDOUT_PREV   =  'mgsmtp_stdout_previousrun.log';
   LOG_FN_SMTP          =  'smtp.log';


constructor TStreamLogger.Create(Stream: TStream);
begin
   LogFile:= Stream;
end;

destructor TStreamLogger.Destroy;
begin
   LogFile.Free;
end;

constructor TLogger.Create(Config: TINIFile);
begin
   FLogOn:= MainServerConfig.Log;
   if FLogOn then begin
      FLogFileName:= Config.ReadString('Log', 'Filename', LOG_FN_SMTP);
      if not FileExists(FLogFileName) then begin
         LogFile:= TFileStream.Create(FLogFileName, fmCreate);
         LogFile.Free;
      end;
      LogFile:= TFileStream.Create(FLogFileName, fmOpenWrite or fmShareDenyWrite);
      LogFile.Seek(0, soFromEnd);
      AddPureLine('');
      AddPureLine('');
      AddPureLine('MgSMTP version ' + MainServerConfig.VersionStr + ', session log');
      AddPureLine('Log started: ' + DateTimeToStr(Now));
   end;
end;

destructor TLogger.Destroy;
begin
   if FLogOn then
      AddPureLine('Log finished: ' + DateTimeToStr(Now));
   inherited Destroy;
end;


function TStreamLogger.AddPureLine(Line: string): boolean;
begin
   Result:= WriteLineToStream(LogFile, Line);
end;

function TStreamLogger.AddLine(Line: string): boolean;
begin
   Result:= AddPureLine('[' + DateTimeToStr(Now) + '] ' + Line);
end;

function TStreamLogger.AddLine(Agent, Line: string): boolean;
begin
   Result:= AddPureLine('[' + DateTimeToStr(Now) + '] <' + Agent + '> ' + Line);
end;

function TStreamLogger.WriteLn(Line: string): boolean;
{ Only a wrapper for AddPureLine for convenience. }
begin
   Result:= AddPureLine(Line);
end;

function TStreamLogger.WriteLn: boolean;
begin
   Result:= AddPureLine('');
end;

function TLogger.AddPureLine(Line: string): boolean;
begin
   if FLogOn then
      Result:= inherited AddPureLine(Line)
   else
      Result:= true;
end;

function TLogger.AddStdLine(Line: string): boolean;
{ Log message to both stdout and smtp.log. }
var outresult, logresult: boolean;
begin
   outresult:= Out.AddLine(Line);
   logresult:= AddLine(Line);
   Result:= outresult and logresult;
end;

function TLogger.AddStdLine(Agent, Line: string): boolean;
{ Log message to both stdout and smtp.log. }
var outresult, logresult: boolean;
begin
   outresult:= Out.AddLine(Agent, Line);
   logresult:= AddLine(Agent, Line);
   Result:= outresult and logresult;
end;



initialization
   if StdOutputHandle > 0 then
      Out:= TStreamLogger.Create(THandleStream.Create(StdOutputHandle))
   else begin
      DeleteFile(LOG_FN_STDOUT_PREV);
      RenameFile(LOG_FN_STDOUT, LOG_FN_STDOUT_PREV);
      Out:= TStreamLogger.Create(TFileStream.Create(LOG_FN_STDOUT, fmCreate or fmShareDenyWrite));
   end;
finalization
   Out.Free;
end.
