{
   Basic object-oriented network functions
   Copyright (C) 2010-2018 MegaBrutal

   This unit is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This unit is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

{
   Unit: Network
   This unit provides an object-oriented interface to manage TCP/IPv4
   connections.

   TTCPConnection - provides methods for sending/receiving buffers through
      the connection, and some support for text-based communication.

   TTCPRFCConnection - in addition to TTCPConnection, it provides methods
      to send/receive RFC-style commands and responses.

   TTCPListener - opens a port to listen on, and accepts incoming connections
      through it. Override its "HandleClient" method to serve connected
      clients.
}


{$MODE DELPHI}
unit Network;

interface
uses Classes, Sockets, SocketUtils, SysUtils, DNSResolve, NetRFC, Common;

const

   { Connection feature requests: }
   NET_TCP_BASIC        =  0;
   NET_TCP_RFCSUPPORT   =  1;

   { Default socket timeout: }
   DEF_SOCK_TIMEOUT     =  5 * 60000;     { 5 minutes. }


type

   TTCPConnection = class
      constructor Create; overload;
      constructor Create(const HostName: string; Port: word); overload;
      constructor Create(Socket: socket); overload;
      destructor Destroy; override;
   private
      FConnected: boolean;
      FSocket: socket;
      FHostIP: TIPNamePair;
      FSockTimeOut: DWord;
      SrcSockAddr, DstSockAddr: TSockAddr;
   public
      function SetBindAddress(const HostName: string): boolean;
      function Connect(const HostName: string; Port: word): boolean;
      procedure Disconnect;
      procedure ReverseDNSLookup;
      function VerifyFCrDNS: boolean;
      procedure SetSockTimeOut(TimeOut: DWord);
      function ReadBuffer(PtrBuffer: pointer; Len: size_t): ssize_t;
      function WriteBuffer(PtrBuffer: pointer; Len: size_t): ssize_t;
      function ReadLn(var Line: string): boolean;
      function WriteLn(const Line: string): boolean;
      property Connected: boolean read FConnected;
      property Socket: socket read FSocket;
      property HostIP: TIPNamePair read FHostIP;
      property SockTimeOut: DWord read FSockTimeOut write SetSockTimeOut;
   end;

   TTCPRFCConnection = class(TTCPConnection)
   public
      function ReadCommand(var Command: shortstring; var Prms: string): boolean;
      function ReadResponse(Response: TRFCReply): boolean;
      function SendCommand(Command: shortstring): boolean; overload;
      function SendCommand(Command: shortstring; Prms: string): boolean; overload;
      function SendResponse(Response: TRFCReply): boolean;
   end;


   TTCPAcceptHandler = procedure(Connection: TTCPConnection) of object;

   TTCPAcceptor = class(TThread)
      constructor Create(Handler: TTCPAcceptHandler; TCPConnection: TTCPConnection);
   protected
      FHandler: TTCPAcceptHandler;
      FTCPConnection: TTCPConnection;
      procedure Execute; override;
   end;

   TTCPListener = class(TThread)
      constructor Create(const Address: string; Port: word; FeatureRequest: word);
      {destructor Destroy; override;}
   protected
      FFeatureRequest: word;
      FListenAddress: string;
      FListenPort: word;
      FListenSocket: socket;
      SockAddr: TSockAddr;
      procedure HandleClient(Connection: TTCPConnection); virtual; abstract;
      procedure Execute; override;
   public
      property ListenAddress: string read FListenAddress;
      property ListenPort: word read FListenPort;
      function GetSockAddrStr: string;
      function StartListen: boolean;
      procedure StopListen;
   end;



implementation


constructor TTCPConnection.Create;
{ Create an instance, but don't connect to anywhere yet. }
begin
   inherited Create;
   FConnected:= false;
   FSocket:= -1;
   FSockTimeOut:= DEF_SOCK_TIMEOUT;
   FillChar(SrcSockAddr, SizeOf(SrcSockAddr), 0);
   FillChar(DstSockAddr, SizeOf(DstSockAddr), 0);
end;

constructor TTCPConnection.Create(const HostName: string; Port: word);
{ Connect to the given port on the given hostname. }
begin
   Create;
   Connect(HostName, Port);
end;

constructor TTCPConnection.Create(Socket: socket);
{ Use an already connected socket. }
var ssocklen, dsocklen: TSockLen;
begin
   inherited Create;
   FSocket:= Socket;
   ssocklen:= SizeOf(SrcSockAddr);
   dsocklen:= SizeOf(DstSockAddr);
   fpgetsockname(FSocket, @SrcSockAddr, @ssocklen);
   fpgetpeername(FSocket, @DstSockAddr, @dsocklen);
   FHostIP:= TIPNamePair.Create('', NetAddrToStr(DstSockAddr.sin_addr));
   FConnected:= true;
end;

destructor TTCPConnection.Destroy;
begin
   if FConnected then Disconnect;
   inherited Destroy;
end;


constructor TTCPAcceptor.Create(Handler: TTCPAcceptHandler; TCPConnection: TTCPConnection);
{ Start a connection handler on a distinct thread. }
begin
   FHandler:= Handler;
   FTCPConnection:= TCPConnection;
   FreeOnTerminate:= true;
   inherited Create(false);
end;


constructor TTCPListener.Create(const Address: string; Port: word; FeatureRequest: word);
begin
   FListenAddress:= Address;
   FListenPort:= Port;
   FFeatureRequest:= FeatureRequest;
   FreeOnTerminate:= false;
   FillChar(SockAddr, SizeOf(SockAddr), 0);
   inherited Create(true);
end;


function TTCPConnection.SetBindAddress(const HostName: string): boolean;
var GAIResult: TGAIResult;
begin
   GAIResult:= ResolveHost(HostName);
   if GAIResult.GAIError = 0 then begin
      SrcSockAddr:= GAIResult.AddrInfo^.ai_addr^;
      FreeHost(GAIResult);
      Result:= true;
   end
   else
      Result:= false;
end;

function TTCPConnection.Connect(const HostName: string; Port: word): boolean;
{ Resolves the given hostname, and tries to connect it on the given port. }
var GAIResult: TGAIResult;
begin
   FSocket:= fpSocket(af_inet, sock_stream, 0);
   if (FSocket <> -1) then begin
      if (SrcSockAddr.sin_addr.s_addr = 0) or (fpBind(FSocket, @SrcSockAddr, SizeOf(SrcSockAddr)) = 0) then begin
         GAIResult:= ResolveHost(HostName);
         if GAIResult.GAIError = 0 then begin
            DstSockAddr:= GAIResult.AddrInfo^.ai_addr^;
            DstSockAddr.sin_port:= htons(Port);

            if DstSockAddr.sin_addr.s_addr <> 0 then
            { Try to initiate connection. }
            FConnected:= fpConnect(FSocket, @DstSockAddr, SizeOf(DstSockAddr)) <> -1;

            if FConnected then begin
               FHostIP:= TIPNamePair.Create(HostName, NetAddrToStr(DstSockAddr.sin_addr));
               SetSockTimeOut(FSockTimeOut);
            end
            else
               CloseSocket(FSocket);

            FreeHost(GAIResult);
         end;
      end;
   end;
   Result:= FConnected;
end;

procedure TTCPConnection.Disconnect;
begin
   fpShutdown(FSocket, 2);
   CloseSocket(FSocket);
   FSocket:= -1;
   FHostIP.Free;
   FConnected:= false;
end;

procedure TTCPConnection.ReverseDNSLookup;
{ Performs a reverse DNS lookup, and updates the HostIP structure. }
var NHostIP: TIPNamePair;
begin
   if FConnected then begin
      NHostIP:= TIPNamePair.Create(ResolveIP(@DstSockAddr), FHostIP.IP);
      FHostIP.Free;
      FHostIP:= NHostIP;
   end;
end;

function TTCPConnection.VerifyFCrDNS: boolean;
var GAIResult: TGAIResult; ai: PAddrInfo;
begin
   Result:= false;
   GAIResult:= ResolveHost(HostIP.Name);
   if GAIResult.GAIError = 0 then begin
      ai:= GAIResult.AddrInfo;
	  { One of the addresses must match. }
	  while (ai <> nil) and not Result do begin
	     Result:= NetAddrToStr(ai^.ai_addr^.sin_addr) = HostIP.IP;
		 ai:= ai^.ai_next;
      end;
   end;
end;

procedure TTCPConnection.SetSockTimeOut(TimeOut: DWord);
begin
   FSockTimeOut:= TimeOut;
   if Connected then begin
      fpSetSockOpt(FSocket, SOL_SOCKET, SO_RCVTIMEO, @FSockTimeOut, SizeOf(FSockTimeOut));
      fpSetSockOpt(FSocket, SOL_SOCKET, SO_SNDTIMEO, @FSockTimeOut, SizeOf(FSockTimeOut));
   end;
end;

function TTCPConnection.ReadBuffer(PtrBuffer: pointer; Len: size_t): ssize_t;
begin
   Result:= fpRecv(FSocket, PtrBuffer, Len, 0);
end;

function TTCPConnection.WriteBuffer(PtrBuffer: pointer; Len: size_t): ssize_t;
begin
   Result:= fpSend(FSocket, PtrBuffer, Len, 0);
end;

function TTCPConnection.ReadLn(var Line: string): boolean;
begin
   Result:= SockReadLn(FSocket, Line);
end;

function TTCPConnection.WriteLn(const Line: string): boolean;
begin
   Result:= SockWriteLn(FSocket, Line);
end;


function TTCPRFCConnection.ReadCommand(var Command: shortstring; var Prms: string): boolean;
begin
   Result:= NetRFC.ReadCommand(FSocket, Command, Prms);
end;

function TTCPRFCConnection.ReadResponse(Response: TRFCReply): boolean;
begin
   Result:= NetRFC.ReadResponse(FSocket, Response);
end;

function TTCPRFCConnection.SendCommand(Command: shortstring): boolean;
begin
   Result:= NetRFC.SendCommand(FSocket, Command);
end;

function TTCPRFCConnection.SendCommand(Command: shortstring; Prms: string): boolean;
begin
   Result:= NetRFC.SendCommand(FSocket, Command, Prms);
end;

function TTCPRFCConnection.SendResponse(Response: TRFCReply): boolean;
begin
   Result:= NetRFC.SendResponse(FSocket, Response);
end;


procedure TTCPAcceptor.Execute;
begin
   FHandler(FTCPConnection);
end;


function TTCPListener.GetSockAddrStr: string;
begin
   Result:= NetAddrToStr(SockAddr.sin_addr) + ':' + IntToStr(ntohs(SockAddr.sin_port));
end;

function TTCPListener.StartListen: boolean;
var GAIResult: TGAIResult;
begin
   FListenSocket:= fpSocket(af_inet, sock_stream, 0);
   if FListenSocket <> -1 then begin
      GAIResult:= ResolveHost(FListenAddress);
	  if GAIResult.GAIError = 0 then begin
	     SockAddr:= GAIResult.AddrInfo^.ai_addr^;
		 SockAddr.sin_port:= htons(FListenPort);

         if fpBind(FListenSocket, @SockAddr, SizeOf(SockAddr)) <> -1 then begin
            { It seems the maximum connection value isn't enforced by the
              Free Pascal library, so this 512 is a constant, dummy value. }
            if fpListen(FListenSocket, 512) <> -1 then begin
               Result:= true;
               Start;
            end
            else Result:= false;
         end
         else Result:= false;

         FreeHost(GAIResult);
      end
	  else Result:= false;
   end
   else Result:= false;
end;

procedure TTCPListener.StopListen;
begin
   Terminate;
   KillThread(Handle);
end;

procedure TTCPListener.Execute;
var ClientSocket: socket; AcceptFailCount: word; Len: longint;
    TCPConnection: TTCPConnection;
begin
   { Now, accept connections. }
   AcceptFailCount:= 0;
   while not Terminated do begin
      Len:= SizeOf(SockAddr);
      ClientSocket:= fpAccept(FListenSocket, @SockAddr, @Len);
      if ClientSocket <> -1 then begin
         AcceptFailCount:= 0;

         { Creates the requested TTCPConnection object for the accepted
           connection. }
         case FFeatureRequest of
         NET_TCP_BASIC:
            TCPConnection:= TTCPConnection.Create(ClientSocket);
         NET_TCP_RFCSUPPORT:
            TCPConnection:= TTCPRFCConnection.Create(ClientSocket);
         end;

         { Then start a new thread with the connection handler. }
         TTCPAcceptor.Create(HandleClient, TCPConnection);
      end
      else begin
         Inc(AcceptFailCount);
         if AcceptFailCount >= 512 then Terminate;
      end;
   end;
end;


end.
