{
   MegaBrutal's SMTP Server (MgSMTP)
   Copyright (C) 2010-2018 MegaBrutal

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
}

{
   Unit: Relay
   This unit implements the necessary objects to relay messages towards
   remote servers. It handles the re-routing of messages, when it's
   configured.
}


{$MODE DELPHI}
unit Relay;

interface
uses SysUtils, Classes, INIFiles, Base64, CompareWild, Common, Network,
     DNSMX, NetRFC, RFCSMTP;

type

   TMailRoute = record
      Mask: string;
      Target: integer;
   end;

   TSMTPExtensions = record
      Pipelining, Size, EbitMIME: boolean;
   end;

   { A TRoutingTarget holds the data for a single relay host.
     The administrator may give a symbolic name to some relay hosts that
     identifies a distinct section in the configuration INI file, where
     all necessary data can be found to contact the particular relay host. }

   TRoutingTarget = class
      constructor Create(Name, TargetHost: string; Port: integer; Auth: boolean; Username, Password: string);
   protected
      FName, FTargetHost, FUsername, FPassword: string;
      FPort: integer;
      FAuth: boolean;
   public
      property Name: string read FName;
      property Host: string read FTargetHost;
      property Port: integer read FPort;
      property Auth: boolean read FAuth;
      property Username: string read FUsername;
      property Password: string read FPassword;
      function Copy: TRoutingTarget;
   end;

   { TRoutingTable manages re-routing. It can be asked where to relay e-mails
     for a specific host. It holds multiple instances of TRoutingTarget. }

   TRoutingTable = class
      constructor Create;
      destructor Destroy; override;
   protected
      Targets: array of TRoutingTarget;
      Routes: array of TMailRoute;
      function FindOrLoadTarget(TargetName, TargetHost: string; Port: integer; Auth: boolean; Username, Password: string): integer;
   public
      procedure AddRoute(Mask: string; TargetName, TargetHost: string; Port: integer; Auth: boolean; Username, Password: string);
      function ReRoute(Host: string): string;
      function GetRouteInfo(Host: string): TRoutingTarget;
   end;

   { TRelayer does the actual relaying to a host. It connects the target server
     and passes the message to it by SMTP protocol. }

   TRelayer = class
      constructor Create(RoutingTable: TRoutingTable; Envelope: TEnvelope; EMailProperties: TEMailProperties);
      destructor Destroy; override;
   protected
      FEnvelope: TEnvelope;
      FEMailProperties: TEMailProperties;
      FTransactionComplete: boolean;
      FRoutingTarget: TRoutingTarget;
      RoutingTable: TRoutingTable;
      TCP: TTCPRFCConnection;
      Response: TRFCReply;
      SMTPExtensions: TSMTPExtensions;
      procedure AdministerMassFailure(var Result: boolean);
      function GetRelayServerName: string;
      function GetRelayServerPort: integer;
   public
      property Envelope: TEnvelope read FEnvelope;
      property EMailProperties: TEMailProperties read FEMailProperties;
      property IsTransactionComplete: boolean read FTransactionComplete;
      property RelayServerName: string read GetRelayServerName;
      property RelayServerPort: integer read GetRelayServerPort;
      function OpenConnection: boolean;
      function Greet: boolean;
      function SendEnvelope: boolean;
      function PrepareSendMessage: boolean;
      function DeliverMessagePart(Chunk: TStrings): boolean;
      procedure FinishDeliverMessage;
      procedure CloseConnection;
   end;

   { TRelayManager is the main manager object of the entire relay unit.
     It loads all the configuration, sets up the corresponding objects,
     and it creates configured TRelayer-s. }

   TRelayManager = class
      constructor Create(Config: TINIFile);
      destructor Destroy; override;
   protected
      RelayToList, NoRelayToList: TStrings;
      RoutingTable: TRoutingTable;
   public
      function CreateRelayer(Envelope: TEnvelope; EMailProperties: TEMailProperties): TRelayer;
      function IsOnRelayToList(HostName: string): boolean;
      function IsOnNoRelayToList(HostName: string): boolean;
      function OrganizeEnvelopes(Envelopes: TEnvelopeArray): TEnvelopeArray;
   end;


var

   RelayManager: TRelayManager;



implementation


constructor TRoutingTarget.Create(Name, TargetHost: string; Port: Integer; Auth: boolean; Username, Password: string);
begin
   inherited Create;
   FName:= Name;
   if TargetHost = '' then FTargetHost:= Name else FTargetHost:= TargetHost;
   FPort:= Port;
   FAuth:= Auth;
   FUsername:= Username;
   FPassword:= Password;
end;

constructor TRoutingTable.Create;
begin
   inherited Create;
   SetLength(Targets, 0);
   SetLength(Routes, 0);
end;

destructor TRoutingTable.Destroy;
var i: integer;
begin
   for i:= 0 to Length(Targets) - 1 do
      Targets[i].Free;
   SetLength(Routes, 0);
   SetLength(Targets, 0);
   inherited Destroy;
end;

constructor TRelayer.Create(RoutingTable: TRoutingTable; Envelope: TEnvelope; EMailProperties: TEMailProperties);
begin
   inherited Create;
   Self.RoutingTable:= RoutingTable;
   FEnvelope:= Envelope;
   FEMailProperties:= EMailProperties;
   FTransactionComplete:= false;
   FRoutingTarget:= RoutingTable.GetRouteInfo(Envelope.RelayHost);
   Response:= TRFCReply.Create;
   FillChar(SMTPExtensions, SizeOf(TSMTPExtensions), #0);
end;

destructor TRelayer.Destroy;
begin
   FRoutingTarget.Free;
   Response.Free;
   inherited Destroy;
end;

constructor TRelayManager.Create(Config: TINIFile);
var i: integer; RouteMasks: TStringList; RouteName: string;
begin
   inherited Create;

   RelayToList:= TStringList.Create;
   RelayToList.Delimiter:= ',';
   RelayToList.DelimitedText:= Config.ReadString('Relay', 'RelayTo', '');

   NoRelayToList:= TStringList.Create;
   NoRelayToList.Delimiter:= ',';
   NoRelayToList.DelimitedText:= Config.ReadString('Relay', 'NoRelayTo', '');

   RoutingTable:= TRoutingTable.Create;
   RouteMasks:= TStringList.Create;
   Config.ReadSection('Relay\Routes', RouteMasks);
   for i:= 0 to RouteMasks.Count - 1 do begin
      RouteName:= Config.ReadString('Relay\Routes', RouteMasks.Strings[i], '!');
      RoutingTable.AddRoute(RouteMasks.Strings[i],
         RouteName,
         Config.ReadString('Relay\Routes\' + RouteName, 'Host', ''),
         Config.ReadInteger('Relay\Routes\' + RouteName, 'Port', STANDARD_SMTP_PORT),
         Config.ReadBool('Relay\Routes\' + RouteName, 'Auth', false),
         Config.ReadString('Relay\Routes\' + RouteName, 'Username', ''),
         Config.ReadString('Relay\Routes\' + RouteName, 'Password', '')
      );
   end;
   RouteMasks.Free;
end;

destructor TRelayManager.Destroy;
begin
   RelayToList.Free;
   RoutingTable.Free;
   inherited Destroy;
end;


function TRoutingTarget.Copy: TRoutingTarget;
begin
   Result:= TRoutingTarget.Create(Name, Host, Port, Auth, Username, Password);
end;

procedure TRoutingTable.AddRoute(Mask: string; TargetName, TargetHost: string; Port: integer; Auth: boolean; Username, Password: string);
{ It should be only called at start-up. It creates the necessary TRountingTarget
  objects. It doesn't create redundant targets. If more entries are there to
  relay to a specific server, then only one TRoutingTarget will be created
  for that relay host. It is ensured by FindOrLoadTarget. }
var i: integer;
begin
   i:= Length(Routes);
   SetLength(Routes, i + 1);
   Routes[i].Mask:= Mask;
   Routes[i].Target:= FindOrLoadTarget(TargetName, TargetHost, Port, Auth, Username, Password);
end;

function TRoutingTable.FindOrLoadTarget(TargetName, TargetHost: string; Port: integer; Auth: boolean; Username, Password: string): integer;
{ Creates a new TRoutingTarget, but only if no other TRoutingTarget exists
  with the same name. If it does find an already-existing TRoutingTarget
  with the given name, it returns that instance. }
var i, x: integer; Found: boolean;
begin
   i:= 0; Found:= false;
   while (i < Length(Targets)) and (not Found) do begin
      if Targets[i].Name = TargetName then begin
         Found:= true;
         x:= i;
      end;
      Inc(i);
   end;
   if not Found then begin
      x:= Length(Targets);
      SetLength(Targets, x + 1);
      Targets[x]:= TRoutingTarget.Create(TargetName, TargetHost, Port, Auth, Username, Password);
   end;
   Result:= x;
end;

function TRoutingTable.ReRoute(Host: string): string;
{ It returns the NAME of the relay host that's supposed to relay messages
  towards the specified host. The mentioned NAME can be a hostname or
  a symbolic name given in the configuration. If this function returns "!",
  that means that the message should be relayed to the named host itself. }
var i: integer; Found: boolean;
begin
   i:= 0; Found:= false;
   while (i < Length(Routes)) and (not Found) do begin
      if WildComp(UpperCase(Routes[i].Mask), UpperCase(Host)) then begin
         Result:= Targets[Routes[i].Target].Name;
         Found:= true;
      end;
      Inc(i);
   end;
   if not Found then Result:= Host;
end;

function TRoutingTable.GetRouteInfo(Host: string): TRoutingTarget;
{ It returns the corresponding TRoutingTarget for a given name.
  That name may be a symbolic name, given in the configuration,
  or a valid hostname.
  Note, this function returns a COPY of the TRoutingTarget.
  The caller is responsible for freeing it.
  If there is no TRoutingTarget with the given name, the function
  creates a new TRoutingTarget and puts the given hostname into it. }
var i: integer; Found: boolean;
begin
   i:= 0; Found:= false;
   while (i < Length(Targets)) and (not Found) do begin
      if Targets[i].Name = Host then begin
         Result:= Targets[i].Copy;
         Found:= true;
      end;
      Inc(i);
   end;
   if not Found then Result:= TRoutingTarget.Create(Host, Host, STANDARD_SMTP_PORT, false, '', '');
end;


procedure TRelayer.AdministerMassFailure(var Result: boolean);
begin
   Envelope.SetAllRecipientData(Response.GetNumericCode, Response.ReplyText.Text);
   Result:= false;
end;

function TRelayer.GetRelayServerName: string;
begin
   Result:= FRoutingTarget.Host;
end;

function TRelayer.GetRelayServerPort: integer;
begin
   Result:= FRoutingTarget.Port;
end;

function TRelayer.OpenConnection: boolean;
{ Initiates connection to the relay site. It queries the MX records for the
  relay site's domain, and tries to connect the resulting hosts in the
  order of MX priorities. If there are no MX records for the domain,
  the domain's A record will be connected.
  The function returns TRUE, if it successfully established connection
  to any of the MX hostnames. }
var MXList: TStrings; i: integer;
begin
   MXList:= GetCorrectMXRecordList(RelayServerName);
   if MXList.Count >= 1 then begin
      TCP:= TTCPRFCConnection.Create;
      TCP.SetBindAddress(MainServerConfig.BindAddress);
      TCP.Connect(MXList.Strings[0], RelayServerPort);
      TCP.SetSockTimeOut(DEF_SOCK_TIMEOUT);
      i:= 1;
      while (not TCP.Connected) and (i < MXList.Count) do begin
         TCP.Connect(MXList.Strings[i], RelayServerPort);
         Inc(i);
      end;
      Result:= TCP.Connected;
   end
   else Result:= false;
   MXList.Free;
   FTransactionComplete:= false;
end;

function TRelayer.Greet: boolean;
{ This function reads and checks the relay server's greeting.
  Then identifies this server with an EHLO.
  Then, if necessary, authenticates at the connected relay server.
  The function returns true, if the authentication and the EHLO command were
  successful. }
var
   i: integer;
   Authenticated: boolean;
   StringStream: TStringStream;
   Base64EncodingStream: TBase64EncodingStream;
   Line: string;

begin
   Response.Clear;
   AdministerMassFailure(Result);
   TCP.ReadResponse(Response);

   { Expect 2xx reply. }
   if (Response.GetNumericCode div 100) = 2 then begin

      TCP.SendCommand(SMTP_C_EHLO, MainServerConfig.Name);
      TCP.ReadResponse(Response);

      if Response.GetNumericCode = SMTP_R_OK then begin
         for i:= 1 to Response.Count - 1 do begin
            Line:= UpperCase(Response.GetLine(i));
            if pos('PIPELINING', Line) = 1 then
               SMTPExtensions.Pipelining:= true
            else if pos('SIZE', Line) = 1 then
               SMTPExtensions.Size:= true
            else if pos('8BITMIME', Line) = 1 then
               SMTPExtensions.EbitMIME:= true;
         end;
         Result:= true;
      end
      else if (Response.GetNumericCode >= 500) and (Response.GetNumericCode <= 504) then begin
         { It seems the remote site did not understand our EHLO, that is,
           let's admit, quite odd in the 21st century...
           Whatever, let's fall back to RFC 821 then. }
         TCP.SendCommand(SMTP_C_HELO, MainServerConfig.Name);
         TCP.ReadResponse(Response);
         Result:= Response.GetNumericCode = SMTP_R_OK;
      end;

      if Result then begin
         if FRoutingTarget.Auth then begin
            TCP.SendCommand(SMTP_C_AUTH, 'LOGIN');
            TCP.ReadResponse(Response);
            if Response.GetNumericCode = SMTP_R_AUTH_MESSAGE then begin
               StringStream:= TStringStream.Create('');
               Base64EncodingStream:= TBase64EncodingStream.Create(StringStream);
               Base64EncodingStream.Write(PChar(FRoutingTarget.Username)^, Length(FRoutingTarget.Username));
               Base64EncodingStream.Destroy;
               TCP.WriteLn(StringStream.DataString);
               StringStream.Destroy;
               TCP.ReadResponse(Response);
               if Response.GetNumericCode = SMTP_R_AUTH_MESSAGE then begin
                  StringStream:= TStringStream.Create('');
                  Base64EncodingStream:= TBase64EncodingStream.Create(StringStream);
                  Base64EncodingStream.Write(PChar(FRoutingTarget.Password)^, Length(FRoutingTarget.Password));
                  Base64EncodingStream.Destroy;
                  TCP.WriteLn(StringStream.DataString);
                  StringStream.Destroy;
                  TCP.ReadResponse(Response);
                  Authenticated:= Response.GetNumericCode = SMTP_R_AUTH_SUCCESSFUL;
               end
               else Authenticated:= false;
            end
            else Authenticated:= false;
         end
         else Authenticated:= true;

         if not Authenticated then AdministerMassFailure(Result);
      end
      else AdministerMassFailure(Result);

   end
   else AdministerMassFailure(Result);
end;

function TRelayer.SendEnvelope: boolean;
{ Sends the envelope (that is the return-path and the recipient addresses).
  The function returns true, if the MAIL command were successful, and the
  relay server has accepted at least one of the recipient addresses.
  This function returns false if a null reply is read, which is considered
  as protocol violation.
  This function is aware of the SMTP extension, named PIPELINING. If it's
  supported by the server, we send RCPT commands stuffed, without waiting
  for a response. After all RCPTs are sent, we check all responses. }
var
   i, c: integer; UltimateFail: boolean; Prms: string;

   procedure ProcessRCPTResponse;
   begin
      TCP.ReadResponse(Response);
      { If we get an "OK" reply code, we increase the count of successful
        recipients. }
      if Response.GetNumericCode = SMTP_R_OK then Inc(c)
      { Response code 0 is non-existent in the SMTP protocol.
        If we receive something we _perceive_ as 0, then it is likely
        that something seriously went wrong. MgSMTP shouldn't treat
        this condition as permanent. }
      else if Response.GetNumericCode = 0 then UltimateFail:= true;
      Envelope.SetRecipientData(i, Response.GetNumericCode, Response.ReplyText.Text);
   end;

begin
   { The MAIL command is considered the beginning of the transaction. }
   FTransactionComplete:= false;
   UltimateFail:= false;
   Response.Clear;
   Prms:= 'FROM:<' + Envelope.ReturnPath + '>';

   if SMTPExtensions.Size then
      Prms:= Prms + ' SIZE=' + IntToStr(EMailProperties.Size);
   if SMTPExtensions.EbitMIME and EMailProperties.HasFlag(EF_8BITMIME) then
      Prms:= Prms + ' BODY=8BITMIME';

   TCP.SendCommand(SMTP_C_MAIL, Prms);
   TCP.ReadResponse(Response);
   if Response.GetNumericCode = SMTP_R_OK then begin
      c:= 0;
      for i:= 0 to Envelope.GetNumberOfRecipients - 1 do begin
         TCP.SendCommand(SMTP_C_RCPT, 'TO:<' + Envelope.GetRecipient(i).Address + '>');
         { If pipelining is not supported, read the responses now. }
         if not SMTPExtensions.Pipelining then ProcessRCPTResponse;
      end;

      { If pipelining is supported, process all responses. }
      if SMTPExtensions.Pipelining then
         for i:= 0 to Envelope.GetNumberOfRecipients - 1 do
            ProcessRCPTResponse;

      Result:= (c <> 0) and (not UltimateFail);

      { If there are no accepted recipients, and no protocol failure is
        discovered, we can't send the DATA command, and practically we
        can do nothing more for this transaction. Therefore it is
        considered complete by protocol. }
      FTransactionComplete:= (c = 0) and (not UltimateFail);

      if not Result then begin
         { Either way, try to reset SMTP state in case of failure. }
         TCP.SendCommand(SMTP_C_RSET);
         TCP.ReadResponse(Response);
      end;
   end
   else AdministerMassFailure(Result);
end;

function TRelayer.PrepareSendMessage;
{ Prepares mail transmission with the DATA command. }
begin
   TCP.SendCommand(SMTP_C_DATA);
   TCP.ReadResponse(Response);
   Result:= Response.GetNumericCode = SMTP_R_START_MAIL_INPUT;
end;

function TRelayer.DeliverMessagePart(Chunk: TStrings): boolean;
{ Sends a chunk of the message. }
var i: integer;
begin
   { Check for lines starting with dots. }
   for i:= 0 to Chunk.Count - 1 do
      if (Length(Chunk.Strings[i]) > 0) and (Chunk.Strings[i][1] = '.') then
         Chunk.Strings[i]:= '.' + Chunk.Strings[i];

   { Send text. }
   Result:= TCP.WriteBuffer(PChar(Chunk.Text), Length(Chunk.Text)) <> -1;
end;

procedure TRelayer.FinishDeliverMessage;
{ Finishes the message with a line containing a single dot. }
var i: integer;
begin
   TCP.WriteLn('.');
   TCP.ReadResponse(Response);

   { Mark the transaction complete, if we have a valid response. }
   FTransactionComplete:= Response.GetNumericCode <> 0;

   for i:= 0 to Envelope.GetNumberOfRecipients - 1 do begin
      { Set status code for recipients those were accepted in the envelope stage: }
      if Envelope.GetRecipient(i).Data = SMTP_R_OK then
         Envelope.SetRecipientData(i, Response.GetNumericCode, Response.ReplyText.Text);
   end;
end;

procedure TRelayer.CloseConnection;
begin
   TCP.SendCommand(SMTP_C_QUIT);
   {TCP.ReadResponse(Response);}
   TCP.Free;
end;


function TRelayManager.CreateRelayer(Envelope: TEnvelope; EMailProperties: TEMailProperties): TRelayer;
begin
   Result:= TRelayer.Create(RoutingTable, Envelope, EMailProperties);
end;

function TRelayManager.IsOnRelayToList(HostName: string): boolean;
begin
   Result:= RelayToList.IndexOf(HostName) <> -1;
end;

function TRelayManager.IsOnNoRelayToList(HostName: string): boolean;
begin
   Result:= NoRelayToList.IndexOf(HostName) <> -1;
end;

function TRelayManager.OrganizeEnvelopes(Envelopes: TEnvelopeArray): TEnvelopeArray;
{ Organizes the given envelopes for relaying.
  This function assumes that input envelopes are containing recipient
  addresses orientating to the same site.
  If it turns out that e-mails for multiple sites must be actually relayed
  through the same relay server, this function merges the envelopes for
  those sites; so later, such e-mails will be transmitted though a single
  connection.

  For example, the configuration file indicates:
  - E-mails for "foo.com" must be relayed through "myrelaysmtp".
  - E-mails for "bar.com" must be also relayed through "myrelaysmtp".
  In this case, the envelopes for "foo.com" and "bar.com" will be merged,
  and the e-mail for these sites will be transmitted in one TCP connection. }

var i, j, k: integer; f: boolean; Recipient: TRecipient; OrgHost, TrgHost: string;
begin
   SetLength(Result, 0);
   for i:= 0 to Length(Envelopes) - 1 do begin
      if Envelopes[i].GetNumberOfRecipients > 0 then begin
         Recipient:= Envelopes[i].GetRecipient(0);
         OrgHost:= EMailHost(Recipient.Address);
         TrgHost:= RoutingTable.ReRoute(OrgHost);
         if TrgHost = '!' then TrgHost:= OrgHost;
         j:= 0; f:= false;
         while (j < Length(Result)) and (not f) do begin
            f:= Result[j].RelayHost = TrgHost;
            Inc(j);
         end;
         { Note, if (not f) then j holds Length(Result). }
         if not f then begin
            SetLength(Result, j + 1);
            Result[j]:= TEnvelope.Create;
            Result[j].ReturnPath:= Envelopes[i].ReturnPath;
            Result[j].RelayHost:= TrgHost;
         end
         else Dec(j); { j must be decremented, because we over-incremented it in the loop. }
         with Result[j] do begin
            { Add first recipient to the envelope. }
            AddRecipient(Recipient);
            { Add the remaining recipients. }
            for k:= 1 to Envelopes[i].GetNumberOfRecipients - 1 do
               AddRecipient(Envelopes[i].GetRecipient(k));
         end;
      end;
   end;
end;


end.
